# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import datetime
import tellme
from .. import compat
from . import helpers

RESULT_CATEGORY_SCHEMA = 'schema'
RESULT_LEVEL_INFO = 'info'
RESULT_LEVEL_WARNING = 'warning'
RESULT_LEVEL_ERROR = 'error'

RESULTS = {
    'incorrect_headers': {
        'id': 'incorrect_headers',
        'name': 'Incorrect Headers',
        'msg': 'The headers do not match the schema.'
    },
    'incorrect_headers': {
        'id': 'incorrect_headers',
        'name': 'Incorrect Headers',
        'msg': 'The headers do not match the schema.'
    },
    'incorrect_headers': {
        'id': 'incorrect_headers',
        'name': 'Incorrect Headers',
        'msg': 'The headers do not match the schema.'
    },
    'incorrect_headers': {
        'id': 'incorrect_headers',
        'name': 'Incorrect Headers',
        'msg': 'The headers do not match the schema.'
    },
    'incorrect_headers': {
        'id': 'incorrect_headers',
        'name': 'Incorrect Headers',
        'msg': 'The headers do not match the schema.'
    },
    'incorrect_headers': {
        'id': 'incorrect_headers',
        'name': 'Incorrect Headers',
        'msg': 'The headers do not match the schema.'
    },
    'incorrect_headers': {
        'id': 'incorrect_headers',
        'name': 'Incorrect Headers',
        'msg': 'The headers do not match the schema.'
    },
    'incorrect_headers': {
        'id': 'incorrect_headers',
        'name': 'Incorrect Headers',
        'msg': 'The headers do not match the schema.'
    },
    'incorrect_headers': {
        'id': 'incorrect_headers',
        'name': 'Incorrect Headers',
        'msg': 'The headers do not match the schema.'
    },
    'invalid_foreignkey_reference_names': {
        'id': 'invalid_foreignkey_reference_names',
        'name': 'Invalid Value for foreignKey Reference',
        'msg': 'A JSON Table Schema foreignKey.reference must match field names.'
    },
    'invalid_foreignkey_reference_value': {
        'id': 'invalid_foreignkey_reference_value',
        'name': 'Invalid Value for foreignKey Reference',
        'msg': 'A JSON Table Schema must have a fields key.'
    },
    'invalid_foreignkey_reference_fields': {
        'id': 'invalid_foreignkey_reference_fields',
        'name': 'Invalid Value for foreignKey Reference',
        'msg': 'A JSON Table Schema foreignKey.reference must have a fields key.'
    },
    'invalid_foreignkey_reference_type': {
        'id': 'invalid_foreignkey_reference_type',
        'name': 'Invalid Type for foreignKey Reference',
        'msg': 'A JSON Table Schema foreignKey.refernce must be a hash.'
    },
    'invalid_foreignkey_reference_value': {
        'id': 'invalid_foreignkey_reference_value',
        'name': 'Invalid Value for foreignKey Reference',
        'msg': 'A JSON Table Schema foreignKey.reference must have a resource key.'
    },
    'invalid_foreignkey_field': {
        'id': 'invalid_foreignkey_field',
        'name': 'Invalid Value in foreignKey Field',
        'msg': 'A JSON Table Schema foreignKey.fields value must correspond with field names.'
    },
    'invalid_foreignkey_fields': {
        'id': 'invalid_foreignkey_fields',
        'name': 'Invalid Value in foreignKey Fields',
        'msg': 'A JSON Table Schema foreignKey.fields value must correspond with field names.'
    },
    'invalid_foreignkey_field_type': {
        'id': 'invalid_foreignkey_field_type',
        'name': 'Invalid Field Type in foreignKey',
        'msg': 'A JSON Table Schema foreignKey.fields type must be a string or an array.'
    },
    'invalid_foreignkey_value': {
        'id': 'invalid_foreignkey_value',
        'name': 'Invalid foreignKey Value',
        'msg': 'A JSON Table Schema foreignKey must have a fields key.'
    },
    'invalid_foreignkey_type': {
        'id': 'invalid_foreignkey_type',
        'name': 'Invalid foreignKey Type',
        'msg': 'A JSON Table Schema foreignKey must be a hash.'
    },
    'invalid_foreignkeys_type': {
        'id': 'invalid_foreignkeys_type',
        'name': 'Invalid foreignKeys Type',
        'msg': 'A JSON Table Schema foreignKeys must be an array.'
    },
    'invalid_primarykey_value': {
        'id': 'invalid_primarykey_value',
        'name': 'Invalid primaryKey Value',
        'msg': 'A JSON Table Schema primaryKey value must be present in the schema field names.'
    },
    'invalid_primarykey_type': {
        'id': 'invalid_primarykey_type',
        'name': 'Invalid primaryKey Type',
        'msg': 'A JSON Table Schema primaryKey must be either a string or an array.'
    },
    'invalid_maximum_constraint_not_supported': {
        'id': 'invalid_maximum_constraint_not_supported',
        'name': 'Invalid Maximum Constraint',
        'msg': 'A JSON Table Schema maximum constraint is present with unclear application (field is not an integer or a date).'
    },
    'invalid_maximum_constraint_datetime': {
        'id': 'invalid_maximum_constraint_datetime',
        'name': 'Invalid Maximum Constraint',
        'msg': 'A JSON Table Schema maximum constraint which is a datetime is only valid if the encompassing field is also a datetime.'
    },
    'invalid_maximum_constraint_time': {
        'id': 'invalid_maximum_constraint_time',
        'name': 'Invalid Maximum Constraint',
        'msg': 'A JSON Table Schema maximum constraint which is a time is only valid if the encompassing field is also a time.'
    },
    'invalid_maximum_constraint_date': {
        'id': 'invalid_maximum_constraint_date',
        'name': 'Invalid Maximum Constraint',
        'msg': 'A JSON Table Schema maximum constraint which is a date is only valid if the encompassing field is also a date.'
    },
    'invalid_maximum_constraint_integer': {
        'id': 'invalid_maximum_constraint_integer',
        'name': 'Invalid Maximum Constraint',
        'msg': 'A JSON Table Schema maximum constraint which is a integer is only valid if the encompassing field is also a integer.'
    },
    'invalid_minimum_constraint_not_supported': {
        'id': 'invalid_minimum_constraint_not_supported',
        'name': 'Invalid Minimum Constraint',
        'msg': 'A JSON Table Schema minimum constraint is present with unclear application (field is not an integer or a date).'
    },
    'invalid_minimum_constraint_datetime': {
        'id': 'invalid_minimum_constraint_datetime',
        'name': 'Invalid Minimum Constraint',
        'msg': 'A JSON Table Schema minimum constraint which is a datetime is only valid if the encompassing field is also a datetime.'
    },
    'invalid_minimum_constraint_time': {
        'id': 'invalid_minimum_constraint_time',
        'name': 'Invalid Minimum Constraint',
        'msg': 'A JSON Table Schema minimum constraint which is a time is only valid if the encompassing field is also a time.'
    },
    'invalid_minimum_constraint_date': {
        'id': 'invalid_minimum_constraint_date',
        'name': 'Invalid Minimum Constraint',
        'msg': 'A JSON Table Schema minimum constraint which is a date is only valid if the encompassing field is also a date.'
    },
    'invalid_minimum_constraint_integer': {
        'id': 'invalid_minimum_constraint_integer',
        'name': 'Invalid Minimum Constraint',
        'msg': 'A JSON Table Schema minimum constraint which is a integer is only valid if the encompassing field is also a integer.'
    },
    'invalid_pattern_constraint': {
        'id': 'invalid_pattern_constraint',
        'name': 'Invalid Pattern Constraint',
        'msg': 'A JSON Table Schema pattern constraint must be a string.'
    },
    'invalid_unique_constraint': {
        'id': 'invalid_unique_constraint',
        'name': 'Invalid Unique Constraint',
        'msg': 'A JSON Table Schema unique constraint must be a boolean.'
    },
    'invalid_maxlength_integer': {
        'id': 'invalid_maxlength_integer',
        'name': 'Invalid maxLength Value',
        'msg': 'A JSON Table Schema maxLength constraint must be an integer.'
    },
    'invalid_minlength_integer': {
        'id': 'invalid_minlength_integer',
        'name': 'Invalid minLength Value',
        'msg': 'A JSON Table Schema minLength constraint must be an integer.'
    },
    'invalid_required_constraint': {
        'id': 'invalid_required_constraint',
        'name': 'Invalid Required Constraint',
        'msg': 'A JSON Table Schema required constraint must be a boolean.'
    },
    'incomplete_field': {
        'id': 'incomplete_field',
        'name': 'Incomplete Field',
        'msg': 'A JSON Table Schema field must have a name key.'
    },
    'invalid_field': {
        'id': 'invalid_field',
        'name': 'Invalid Field',
        'msg': 'Each field in a JSON Table Schema must be a hash.'
    },
    'invalid_field_constraints': {
        'id': 'invalid_field_constraints',
        'name': 'Invalid Field Constraints',
        'msg': 'A JSON Table Schema field constraint must be a hash.'
    },
    'invalid_fields': {
        'id': 'invalid_fields',
        'name': 'Invalid Fields',
        'msg': 'A JSON Table Schema must have an array of fields.'
    },
    'invalid_schema': {
        'id': 'invalid_schema',
        'name': 'Invalid Schema',
        'msg': 'A JSON Table Schema must be a hash.'
    },
    'missing_fields': {
        'id': 'missing_fields',
        'name': 'Missing Fields',
        'msg': 'A JSON Table Schema must have a fields key.'
    },
    'invalid_foreignkey_reference_value': {
        'id': 'invalid_foreignkey_reference_value',
        'name': 'Invalid Value for foreignKey Reference',
        'msg': 'A JSON Table Schema must have a fields key.'
    },
}


def validate(schema, report=None):

    """Validate that `schema` is a valid JSON Table Schema.

    Args:
    * `schema`: a dict to check if it is valid JSON Table Schema
    * `report`: a reporter.Report instance, otherwise a new one is created.

    Returns:
    * A tuple of `valid`, `report`

    TODO: Tidy it all up. test each condition.

    """

    valid = True

    if report is None:
        report = tellme.Report(schema=helpers.report_schema)

    # a schema is a hash
    if not isinstance(schema, dict):

        valid = False
        _type = RESULTS['invalid_schema']
        entry = {
            'processor': 'schema',
            'result_category': RESULT_CATEGORY_SCHEMA,
            'result_level': RESULT_LEVEL_ERROR,
            'result_message': _type['msg'],
            'result_id': _type['id'],
            'result_name': _type['name'],
            'row_index': None,
            'row_name': '',
            'column_index': None,
            'column_name': ''
        }

        report.write(entry)

        # return early in this case.
        return valid, report

    # which MUST contain a key `fields`
    if not schema.get('fields'):

        valid = False
        _type = RESULTS['missing_fields']
        entry = {
            'processor': 'schema',
            'result_category': RESULT_CATEGORY_SCHEMA,
            'result_level': RESULT_LEVEL_ERROR,
            'result_message': _type['msg'],
            'result_id': _type['id'],
            'result_name': _type['name'],
            'row_index': None,
            'row_name': '',
            'column_index': None,
            'column_name': ''
        }

        report.write(entry)

        # return early in this case.
        return valid, report

    # `fields` MUST be an array
    if not isinstance(schema.get('fields'), list):

        valid = False
        _type = RESULTS['invalid_fields']
        entry = {
            'processor': 'schema',
            'result_category': RESULT_CATEGORY_SCHEMA,
            'result_level': RESULT_LEVEL_ERROR,
            'result_message': _type['msg'],
            'result_id': _type['id'],
            'result_name': _type['name'],
            'row_index': None,
            'row_name': '',
            'column_index': None,
            'column_name': ''
        }

        report.write(entry)

        # return early in this case
        return valid, report

    # each entry in the `fields` array MUST be a hash
    if not all([isinstance(o, dict) for o in schema['fields']]):

        valid = False
        _type = RESULTS['invalid_field']
        entry = {
            'processor': 'schema',
            'result_category': RESULT_CATEGORY_SCHEMA,
            'result_level': RESULT_LEVEL_ERROR,
            'result_message': _type['msg'],
            'result_id': _type['id'],
            'result_name': _type['name'],
            'row_index': None,
            'row_name': '',
            'column_index': None,
            'column_name': ''
        }

        report.write(entry)

    # each entry in the `fields` array MUST have a `name` key
    if not all([o.get('name') for o in schema['fields']]):

        valid = False
        _type = RESULTS['incomplete_field']
        entry = {
            'processor': 'schema',
            'result_category': RESULT_CATEGORY_SCHEMA,
            'result_level': RESULT_LEVEL_ERROR,
            'result_message': _type['msg'],
            'result_id': _type['id'],
            'result_name': _type['name'],
            'row_index': None,
            'row_name': '',
            'column_index': None,
            'column_name': ''
        }

        report.write(entry)

    # each entry in the `fields` array MAY have a `constraints` key
    # if `constraints` is present, then `constraints` MUST be a hash
    if not all([isinstance(o['constraints'], dict) for o in
               schema['fields'] if o.get('constraints')]):

        valid = False
        _type = RESULTS['invalid_field_constraints']
        entry = {
            'processor': 'schema',
            'result_category': RESULT_CATEGORY_SCHEMA,
            'result_level': RESULT_LEVEL_ERROR,
            'result_message': _type['msg'],
            'result_id': _type['id'],
            'result_name': _type['name'],
            'row_index': None,
            'row_name': '',
            'column_index': None,
            'column_name': ''
        }

        report.write(entry)

    # constraints may contain certain keys (each has a specific meaning)
    for constraints, _type in [(o['constraints'], o.get('type'))
                               for o in schema['fields'] if
                               o.get('constraints')]:

        # IF `required` key, then it is a boolean
        if constraints.get('required') and not \
                isinstance(constraints['required'], bool):

            valid = False
            _type = RESULTS['invalid_required_constraint']
            entry = {
                'processor': 'schema',
                'result_category': RESULT_CATEGORY_SCHEMA,
                'result_level': RESULT_LEVEL_ERROR,
                'result_message': _type['msg'],
                'result_id': _type['id'],
                'result_name': _type['name'],
                'row_index': None,
                'row_name': '',
                'column_index': None,
                'column_name': ''
            }

            report.write(entry)

        # IF `minLength` key, then it is an integer
        if constraints.get('minLength') and not \
                isinstance(constraints['minLength'], int):

            valid = False
            _type = RESULTS['invalid_minlength_integer']
            entry = {
                'processor': 'schema',
                'result_category': RESULT_CATEGORY_SCHEMA,
                'result_level': RESULT_LEVEL_ERROR,
                'result_message': _type['msg'],
                'result_id': _type['id'],
                'result_name': _type['name'],
                'row_index': None,
                'row_name': '',
                'column_index': None,
                'column_name': ''
            }

            report.write(entry)

        # IF `maxLength` key, then it is an integer
        if constraints.get('maxLength') and not \
                isinstance(constraints['maxLength'], int):

            valid = False
            _type = RESULTS['invalid_maxlength_integer']
            entry = {
                'processor': 'schema',
                'result_category': RESULT_CATEGORY_SCHEMA,
                'result_level': RESULT_LEVEL_ERROR,
                'result_message': _type['msg'],
                'result_id': _type['id'],
                'result_name': _type['name'],
                'row_index': None,
                'row_name': '',
                'column_index': None,
                'column_name': ''
            }

            report.write(entry)

        # IF `unique` key, then it is a boolean
        if constraints.get('unique') and not \
                isinstance(constraints['unique'], bool):

            valid = False
            _type = RESULTS['invalid_unique_constraint']
            entry = {
                'processor': 'schema',
                'result_category': RESULT_CATEGORY_SCHEMA,
                'result_level': RESULT_LEVEL_ERROR,
                'result_message': _type['msg'],
                'result_id': _type['id'],
                'result_name': _type['name'],
                'row_index': None,
                'row_name': '',
                'column_index': None,
                'column_name': ''
            }

            report.write(entry)

        # IF `pattern` key, then it is a regex
        if constraints.get('pattern') and not \
                isinstance(constraints['pattern'], compat.str):

            valid = False
            _type = RESULTS['invalid_pattern_constraint']
            entry = {
                'processor': 'schema',
                'result_category': RESULT_CATEGORY_SCHEMA,
                'result_level': RESULT_LEVEL_ERROR,
                'result_message': _type['msg'],
                'result_id': _type['id'],
                'result_name': _type['name'],
                'row_index': None,
                'row_name': '',
                'column_index': None,
                'column_name': ''
            }

            report.write(entry)

        # IF `minimum` key, then it DEPENDS on `field` TYPE
        if constraints.get('minimum'):

            # IF `type` is integer
            if isinstance(_type, int) and not \
                    isinstance(constraints['minimum'], int):

                valid = False
                _type = RESULTS['invalid_minimum_constraint_integer']
                entry = {
                    'processor': 'schema',
                    'result_category': RESULT_CATEGORY_SCHEMA,
                    'result_level': RESULT_LEVEL_ERROR,
                    'result_message': _type['msg'],
                    'result_id': _type['id'],
                    'result_name': _type['name'],
                    'row_index': None,
                    'row_name': '',
                    'column_index': None,
                    'column_name': ''
                }

                report.write(entry)

            # IF `type` is date
            elif isinstance(_type, datetime.date) and not \
                    isinstance(constraints['minimum'], datetime.date):

                valid = False
                _type = RESULTS['invalid_minimum_constraint_date']
                entry = {
                    'processor': 'schema',
                    'result_category': RESULT_CATEGORY_SCHEMA,
                    'result_level': RESULT_LEVEL_ERROR,
                    'result_message': _type['msg'],
                    'result_id': _type['id'],
                    'result_name': _type['name'],
                    'row_index': None,
                    'row_name': '',
                    'column_index': None,
                    'column_name': ''
                }

                report.write(entry)

            # IF `type` is time
            elif isinstance(_type, datetime.time) and not \
                    isinstance(constraints['minimum'], datetime.time):

                valid = False
                _type = RESULTS['invalid_minimum_constraint_time']
                entry = {
                    'processor': 'schema',
                    'result_category': RESULT_CATEGORY_SCHEMA,
                    'result_level': RESULT_LEVEL_ERROR,
                    'result_message': _type['msg'],
                    'result_id': _type['id'],
                    'result_name': _type['name'],
                    'row_index': None,
                    'row_name': '',
                    'column_index': None,
                    'column_name': ''
                }

                report.write(entry)

            # IF `type` is datetime
            elif isinstance(_type, datetime.datetime) and not \
                    isinstance(constraints['minimum'], datetime.datetime):

                valid = False
                _type = RESULTS['invalid_minimum_constraint_datetime']
                entry = {
                    'processor': 'schema',
                    'result_category': RESULT_CATEGORY_SCHEMA,
                    'result_level': RESULT_LEVEL_ERROR,
                    'result_message': _type['msg'],
                    'result_id': _type['id'],
                    'result_name': _type['name'],
                    'row_index': None,
                    'row_name': '',
                    'column_index': None,
                    'column_name': ''
                }

                report.write(entry)

            else:

                valid = False
                _type = RESULTS['invalid_minimum_constraint_not_supported']
                entry = {
                    'processor': 'schema',
                    'result_category': RESULT_CATEGORY_SCHEMA,
                    'result_level': RESULT_LEVEL_ERROR,
                    'result_message': _type['msg'],
                    'result_id': _type['id'],
                    'result_name': _type['name'],
                    'row_index': None,
                    'row_name': '',
                    'column_index': None,
                    'column_name': ''
                }

                report.write(entry)

        # IF `maximum` key, then it DEPENDS on `field` TYPE
        if constraints.get('maximum'):

            # IF `type` is integer
            if isinstance(_type, int) and not \
                    isinstance(constraints['maximum'], int):

                valid = False
                _type = RESULTS['invalid_maximum_constraint_integer']
                entry = {
                    'processor': 'schema',
                    'result_category': RESULT_CATEGORY_SCHEMA,
                    'result_level': RESULT_LEVEL_ERROR,
                    'result_message': _type['msg'],
                    'result_id': _type['id'],
                    'result_name': _type['name'],
                    'row_index': None,
                    'row_name': '',
                    'column_index': None,
                    'column_name': ''
                }

                report.write(entry)

            # IF `type` is date
            elif isinstance(_type, datetime.date) and not \
                    isinstance(constraints['maximum'], datetime.date):

                valid = False
                _type = RESULTS['invalid_maximum_constraint_date']
                entry = {
                    'processor': 'schema',
                    'result_category': RESULT_CATEGORY_SCHEMA,
                    'result_level': RESULT_LEVEL_ERROR,
                    'result_message': _type['msg'],
                    'result_id': _type['id'],
                    'result_name': _type['name'],
                    'row_index': None,
                    'row_name': '',
                    'column_index': None,
                    'column_name': ''
                }

                report.write(entry)

            # IF `type` is time
            elif isinstance(_type, datetime.time) and not \
                    isinstance(constraints['maximum'], datetime.time):

                valid = False
                _type = RESULTS['invalid_maximum_constraint_time']
                entry = {
                    'processor': 'schema',
                    'result_category': RESULT_CATEGORY_SCHEMA,
                    'result_level': RESULT_LEVEL_ERROR,
                    'result_message': _type['msg'],
                    'result_id': _type['id'],
                    'result_name': _type['name'],
                    'row_index': None,
                    'row_name': '',
                    'column_index': None,
                    'column_name': ''
                }

                report.write(entry)

            # IF `type` is datetime
            elif isinstance(_type, datetime.datetime) and not \
                    isinstance(constraints['maximum'], datetime.datetime):

                valid = False
                _type = RESULTS['invalid_maximum_constraint_datetime']
                entry = {
                    'processor': 'schema',
                    'result_category': RESULT_CATEGORY_SCHEMA,
                    'result_level': RESULT_LEVEL_ERROR,
                    'result_message': _type['msg'],
                    'result_id': _type['id'],
                    'result_name': _type['name'],
                    'row_index': None,
                    'row_name': '',
                    'column_index': None,
                    'column_name': ''
                }

                report.write(entry)

            else:

                valid = False
                _type = RESULTS['invalid_maximum_constraint_not_supported']
                entry = {
                    'processor': 'schema',
                    'result_category': RESULT_CATEGORY_SCHEMA,
                    'result_level': RESULT_LEVEL_ERROR,
                    'result_message': _type['msg'],
                    'result_id': _type['id'],
                    'result_name': _type['name'],
                    'row_index': None,
                    'row_name': '',
                    'column_index': None,
                    'column_name': ''
                }

                report.write(entry)

    # the hash MAY contain a key `primaryKey`
    if schema.get('primaryKey'):

        # `primaryKey` MUST be a string or an array
        if not isinstance(schema['primaryKey'], (compat.str, list)):

            valid = False
            _type = RESULTS['invalid_primarykey_type']
            entry = {
                'processor': 'schema',
                'result_category': RESULT_CATEGORY_SCHEMA,
                'result_level': RESULT_LEVEL_ERROR,
                'result_message': _type['msg'],
                'result_id': _type['id'],
                'result_name': _type['name'],
                'row_index': None,
                'row_name': '',
                'column_index': None,
                'column_name': ''
            }

            report.write(entry)

        # ensure that the primary key matches field names
        if isinstance(schema['primaryKey'], compat.str):
            if not schema['primaryKey'] in [f['name'] for f in schema['fields']]:

                valid = False
                _type = RESULTS['invalid_primarykey_value']
                entry = {
                    'processor': 'schema',
                    'result_category': RESULT_CATEGORY_SCHEMA,
                    'result_level': RESULT_LEVEL_ERROR,
                    'result_message': _type['msg'],
                    'result_id': _type['id'],
                    'result_name': _type['name'],
                    'row_index': None,
                    'row_name': '',
                    'column_index': None,
                    'column_name': ''
                }

                report.write(entry)

        else:
            for k in schema['primaryKey']:
                if not k in [f['name'] for f in schema['fields']]:

                    valid = False
                    _type = RESULTS['invalid_primarykey_value']
                    entry = {
                        'processor': 'schema',
                        'result_category': RESULT_CATEGORY_SCHEMA,
                        'result_level': RESULT_LEVEL_ERROR,
                        'result_message': _type['msg'],
                        'result_id': _type['id'],
                        'result_name': _type['name'],
                        'row_index': None,
                        'row_name': '',
                        'column_index': None,
                        'column_name': ''
                    }

                    report.write(entry)

    # the hash may contain a key `foreignKeys`
    if schema.get('foreignKeys'):

        # `foreignKeys` MUST be an array
        if not isinstance(schema['foreignKeys'], list):

            valid = False
            _type = RESULTS['invalid_foreignkeys_type']
            result = {
                'processor': 'schema',
                'result_category': RESULT_CATEGORY_SCHEMA,
                'result_level': RESULT_LEVEL_ERROR,
                'result_message': _type['msg'],
                'result_id': _type['id'],
                'result_name': _type['name'],
                'row_index': None,
                'row_name': '',
                'column_index': None,
                'column_name': ''
            }

            report.write(entry)

        # each `foreignKey` in `foreignKeys` MUST be a hash
        if not all([isinstance(o, dict) for o in
                   schema['foreignKeys']]):

            valid = False
            _type = RESULTS['invalid_foreignkey_type']
            entry = {
                'processor': 'schema',
                'result_category': RESULT_CATEGORY_SCHEMA,
                'result_level': RESULT_LEVEL_ERROR,
                'result_message': _type['msg'],
                'result_id': _type['id'],
                'result_name': _type['name'],
                'row_index': None,
                'row_name': '',
                'column_index': None,
                'column_name': ''
            }

            report.write(entry)

        # each `foreignKey` in `foreignKeys` MUST have a `fields` key
        if not all([o.get('fields') for o in
                   schema['foreignKeys']]):

            valid = False
            _type = RESULTS['invalid_foreignkey_value']
            entry = {
                'processor': 'schema',
                'result_category': RESULT_CATEGORY_SCHEMA,
                'result_level': RESULT_LEVEL_ERROR,
                'result_message': _type['msg'],
                'result_id': _type['id'],
                'result_name': _type['name'],
                'row_index': None,
                'row_name': '',
                'column_index': None,
                'column_name': ''
            }

            report.write(entry)

        # each `fields` key in a `foreignKey` MUST be a string or array
        if not all([isinstance(o.get('fields'), (compat.str, list))
                   for o in schema['foreignKeys']]):

            valid = False
            _type = RESULTS['invalid_foreignkey_field_type']
            entry = {
                'processor': 'schema',
                'result_category': RESULT_CATEGORY_SCHEMA,
                'result_level': RESULT_LEVEL_ERROR,
                'result_message': _type['msg'],
                'result_id': _type['id'],
                'result_name': _type['name'],
                'row_index': None,
                'row_name': '',
                'column_index': None,
                'column_name': ''
            }

            report.write(entry)

        for fk in schema['foreignKeys']:

            # ensure that `foreignKey.fields` match field names
            if isinstance(fk.get('fields'), compat.str):
                if fk.get('fields') not in [f['name'] for f in
                                            schema['fields']]:

                    valid = False
                    _type = RESULTS['invalid_foreignkey_fields']
                    entry = {
                        'processor': 'schema',
                        'result_category': RESULT_CATEGORY_SCHEMA,
                        'result_level': RESULT_LEVEL_ERROR,
                        'result_message': _type['msg'],
                        'result_id': _type['id'],
                        'result_name': _type['name'],
                        'row_index': None,
                        'row_name': '',
                        'column_index': None,
                        'column_name': ''
                    }

                    report.write(entry)

            else:
                for field in fk.get('fields'):
                    if not field in [f['name'] for f in
                                     schema['fields']]:

                        valid = False
                        _type = RESULTS['invalid_foreignkey_field']
                        entry = {
                            'processor': 'schema',
                            'result_category': RESULT_CATEGORY_SCHEMA,
                            'result_level': RESULT_LEVEL_ERROR,
                            'result_message': _type['msg'],
                            'result_id': _type['id'],
                            'result_name': _type['name'],
                            'row_index': None,
                            'row_name': '',
                            'column_index': None,
                            'column_name': ''
                        }

                        report.write(entry)

            # ensure that `foreignKey.reference` is present and is a hash
            if not isinstance(fk.get('reference'), dict):

                valid = False
                _type = RESULTS['invalid_foreignkey_reference_type']
                entry = {
                    'processor': 'schema',
                    'result_category': RESULT_CATEGORY_SCHEMA,
                    'result_level': RESULT_LEVEL_ERROR,
                    'result_message': _type['msg'],
                    'result_id': _type['id'],
                    'result_name': _type['name'],
                    'row_index': None,
                    'row_name': '',
                    'column_index': None,
                    'column_name': ''
                }

                report.write(entry)

            # ensure that `foreignKey.reference` has a `resource` key
            if not 'resource' in fk.get('reference', {}):

                valid = False
                _type = RESULTS['invalid_foreignkey_reference_value']
                entry = {
                    'processor': 'schema',
                    'result_category': RESULT_CATEGORY_SCHEMA,
                    'result_level': RESULT_LEVEL_ERROR,
                    'result_message': _type['msg'],
                    'result_id': _type['id'],
                    'result_name': _type['name'],
                    'row_index': None,
                    'row_name': '',
                    'column_index': None,
                    'column_name': ''
                }

                report.write(entry)

            # ensure that `foreignKey.reference` has a `fields` key
            if not 'fields' in fk.get('reference', {}):

                valid = False
                _type = RESULTS['invalid_foreignkey_reference_fields']
                entry = {
                    'processor': 'schema',
                    'result_category': RESULT_CATEGORY_SCHEMA,
                    'result_level': RESULT_LEVEL_ERROR,
                    'result_message': _type['msg'],
                    'result_id': _type['id'],
                    'result_name': _type['name'],
                    'row_index': None,
                    'row_name': '',
                    'column_index': None,
                    'column_name': ''
                }

                report.write(entry)

            # ensure that `foreignKey.reference.fields`
            # matches outer `fields`
            if isinstance(fk.get('fields'), compat.str):
                if not isinstance(fk['reference']['fields'], compat.str):

                    valid = False
                    _type = RESULTS['invalid_foreignkey_reference_names']
                    entry = {
                        'processor': 'schema',
                        'result_category': RESULT_CATEGORY_SCHEMA,
                        'result_level': RESULT_LEVEL_ERROR,
                        'result_message': _type['msg'],
                        'result_id': _type['id'],
                        'result_name': _type['name'],
                        'row_index': None,
                        'row_name': '',
                        'column_index': None,
                        'column_name': ''
                    }

                    report.write(entry)

            else:
                if not len(fk.get('fields')) == len(fk['reference']['fields']):

                    valid = False
                    _type = RESULTS['invalid_foreignkey_reference_value']
                    entry = {
                        'processor': 'schema',
                        'result_category': RESULT_CATEGORY_SCHEMA,
                        'result_level': RESULT_LEVEL_ERROR,
                        'result_message': _type['msg'],
                        'result_id': _type['id'],
                        'result_name': _type['name'],
                        'row_index': None,
                        'row_name': '',
                        'column_index': None,
                        'column_name': ''
                    }

                    report.write(entry)

    return valid, report



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from . import data_table
from . import table_schema
from . import csv_dialect
from . import data_package
from . import helpers


__all__ = ['data_table', 'table_schema', 'csv_dialect', 'data_package',
           'helpers']



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from . import pipeline
from ..utilities import data_table
from ..utilities import helpers


class Batch(object):

    """Run a validation pipeline batch process."""

    def __init__(self, source, source_type='csv', data_key='data',
                 schema_key='schema', pipeline_options=None,
                 post_task=None, pipeline_post_task=None):

        # `csv` and `dp` (data package) as supported source types
        # data_key is required, but schema_key is not

        self.source = source
        self.source_type = source_type
        self.data_key = data_key
        self.schema_key = schema_key
        self.dataset = self.get_dataset()
        self.pipeline_options = pipeline_options or {}
        self.post_task = post_task
        self.pipeline_post_task = pipeline_post_task
        self.pipeline = None

    def get_dataset(self):
        """Get the dataset for this batch process."""

        return getattr(self, 'get_dataset_{0}'.format(self.source_type))()

    def get_dataset_csv(self):
        """Get the dataset from a CSV file for this batch process."""

        dataset = []
        resources = data_table.DataTable(self.source)

        data_index = resources.headers.index(self.data_key)
        schema_index = None
        if self.schema_key:
            schema_index = resources.headers.index(self.schema_key)

        for entry in resources.values:

            rv = {'data': entry[data_index], 'schema': None}

            if schema_index is not None:
                rv['schema'] = entry[schema_index]

            dataset.append(rv)

        return dataset

    def get_dataset_dp(self):
        """Get the dataset from a Data Package for this batch process."""

        dataset = []
        dp = helpers.load_json_source(self.source)

        for entry in dp['resources']:
            # TODO: need to place the path inside the content of the dp: i.e.: the URL
            rv = {'data': entry['path'], 'schema': entry.get('schema')}

            dataset.append(rv)

        return dataset

    def pipeline_factory(self, data, schema):
        """Construct a pipeline."""

        options = self.pipeline_options or {}
        if options.get('schema') is None:
            options['schema'] = {}
        options['schema']['schema'] = schema

        return pipeline.Pipeline(data, options=options,
                                 post_task=self.pipeline_post_task)

    def run(self):
        """Run the batch."""

        # TODO: parallelize

        for data in self.dataset:
            pipeline = self.pipeline_factory(data['data'], data['schema'])
            pipeline.run()

        if self.post_task:
            # TODO: handle errors from here, etc.
            self.post_task(self)

        return True



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import jtskit
from . import base


RESULTS = {
    'incorrect_headers': {
        'id': 'incorrect_headers',
        'name': 'Incorrect Headers',
        'msg': 'The headers do not match the schema. Data has {0}, but they should be {1}.',
        'help': '',
        'help_edit': ''
    },
    'incorrect_dimensions': {
        'id': 'incorrect_dimensions',
        'name': 'Incorrect Dimensions',
        'msg': 'The row dimensions do not match the header dimensions.',
        'help': '',
        'help_edit': ''
    },
    'incorrect_type': {
        'id': 'incorrect_type',
        'name': 'Incorrect Type',
        'msg': 'The value is not a valid {0}.',
        'help': '',
        'help_edit': ''
    },
    'required_field': {
        'id': 'required_field',
        'name': 'Required Field',
        'msg': 'Column {0} is a required field, but no value can be found in row {1}.',
        'help': '',
        'help_edit': ''
    },
    'nonrequired_field_empty': {
        'id': 'nonrequired_field_empty',
        'name': 'Non-Required Field (Empty/Null)',
        'msg': 'Column {0} is a non-required field, and has a null value in row {1}.',
        'help': '',
        'help_edit': ''
    }
}


class SchemaProcessor(base.Processor):

    """Process data against a JSON Table Schema."""

    name = 'schema'
    RESULT_TYPES = RESULTS

    def __init__(self, fail_fast=False, transform=False, report_limit=1000,
                 row_limit=30000, schema=None, ignore_field_order=True,
                 report_stream=None, report=None, result_level='error',
                 **kwargs):

        super(SchemaProcessor, self).__init__(
            fail_fast=fail_fast, transform=transform,
            report_limit=report_limit, row_limit=row_limit,
            report_stream=report_stream, report=report, result_level=result_level)

        self.ignore_field_order = ignore_field_order
        if not schema:
            self.schema = None
        else:
            self.schema = self.schema_model(schema)

    def schema_model(self, schema):
        return jtskit.models.JSONTableSchema(schema)

#    def pre_run(self, data_table):
#        if self.schema is None:
#            # make a schema
#            # TODO: 50 here is arbitrary
#            sample_data = [row for row in data_table.values][:50]
#            guessed_schema = table_schema.make(data_table.headers, sample_data)
#            self.schema = self.schema_model(guessed_schema)
#
#        return True, data_table

    def run_header(self, headers, header_index=0):

        valid = True

        if self.schema:
            if self.ignore_field_order:
                if not (set(headers).issuperset(set(self.schema.required_headers))):

                    valid = False
                    _type = RESULTS['incorrect_headers']
                    entry = self.make_entry(
                        self.name,
                        self.RESULT_CATEGORY_HEADER,
                        self.RESULT_LEVEL_ERROR,
                        _type['msg'].format(headers, self.schema.headers),
                        _type['id'],
                        _type['name'],
                        headers,
                        header_index,
                        self.RESULT_HEADER_ROW_NAME
                    )

                    self.report.write(entry)
                    if self.fail_fast:
                        return valid, headers

            else:
                header_length = len(headers)
                if not (headers == self.schema.required_headers[:header_length]):

                    valid = False
                    _type = RESULTS['incorrect_headers']
                    entry = self.make_entry(
                        self.name,
                        self.RESULT_CATEGORY_HEADER,
                        self.RESULT_LEVEL_ERROR,
                        _type['msg'].format(headers, self.schema.headers),
                        _type['id'],
                        _type['name'],
                        headers,
                        header_index,
                        self.RESULT_HEADER_ROW_NAME,
                    )

                    self.report.write(entry)
                    if self.fail_fast:
                        return valid, headers

        return valid, headers

    def run_row(self, headers, index, row):

        valid = True
        row_name = self.get_row_id(headers, row)

        if self.schema:
            if not (len(headers) == len(row)):

                valid = False
                _type = RESULTS['incorrect_dimensions']
                entry = self.make_entry(
                    self.name,
                    self.RESULT_CATEGORY_ROW,
                    self.RESULT_LEVEL_ERROR,
                    _type['msg'],
                    _type['id'],
                    _type['name'],
                    row,
                    index,
                    row_name,
                )

                self.report.write(entry)
                if self.fail_fast:
                    return valid, headers, index, row

            else:
                for column_name, column_value in zip(headers, row):
                    # check type and format
                    if self.schema.cast(column_name, column_value) is False:

                        valid = False
                        _type = RESULTS['incorrect_type']
                        entry = self.make_entry(
                            self.name,
                            self.RESULT_CATEGORY_ROW,
                            self.RESULT_LEVEL_ERROR,
                            _type['msg'].format(self.schema.get_type(column_name).name.title()),
                            _type['id'],
                            _type['name'],
                            row,
                            index,
                            row_name,
                            headers.index(column_name),
                            column_name
                        )

                        self.report.write(entry)
                        if self.fail_fast:
                            return valid, headers, index, row

                    # CONSTRAINTS
                    constraints = self.schema.get_constraints(column_name)

                    if constraints['required'] is True and \
                       (column_value in self.schema.NULL_VALUES):

                        valid = False
                        _type = RESULTS['required_field']
                        entry = self.make_entry(
                            self.name,
                            self.RESULT_CATEGORY_ROW,
                            self.RESULT_LEVEL_ERROR,
                            _type['msg'].format(column_name, index),
                            _type['id'],
                            _type['name'],
                            row,
                            index,
                            row_name,
                            headers.index(column_name),
                            column_name
                            )

                        self.report.write(entry)
                        if self.fail_fast:
                            return valid, headers, index, row

                    if constraints['required'] is False and \
                           (column_value in self.schema.NULL_VALUES) and \
                           self.result_level == self.RESULT_LEVEL_INFO:
                        # add info result
                        _type = RESULTS['nonrequired_field_empty']
                        entry = self.make_entry(
                            self.name,
                            self.RESULT_CATEGORY_ROW,
                            self.RESULT_LEVEL_INFO,
                            _type['msg'].format(column_name, index),
                            _type['id'],
                            _type['name'],
                            row,
                            index,
                            row_name,
                            headers.index(column_name),
                            column_name
                        )

                        self.report.write(entry)

                    # TODO: check constraints.unique
                    # TODO: check constraints.min* and constraints.max*

        return valid, headers, index, row



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import copy
from . import base


RESULTS = {
    'empty_header': {
        'id': 'empty_header',
        'name': 'Empty Header',
        'msg': 'The header in column {0} was found to be empty.',
        'help': '',
        'help_edit': ''
    },
    'duplicate_header': {
        'id': 'duplicate_header',
        'name': 'Duplicate Header',
        'msg': 'The header in column {0} was found to have duplicates.',
        'help': '',
        'help_edit': ''
    },
    'defective_row': {
        'id': 'defective_row',
        'name': 'Defective Row',
        'msg': 'Row {0} is defective: the dimensions are incorrect compared to headers.',
        'help': '',
        'help_edit': ''
    },
    'duplicate_row': {
        'id': 'duplicate_row',
        'name': 'Duplicate Row',
        'msg': 'Row {0} duplicates the following rows which have already been seen: {1}.',
        'help': '',
        'help_edit': ''
    },
    'empty_row': {
        'id': 'empty_row',
        'name': 'Empty Row',
        'msg': 'Row {0} is empty.',
        'help': '',
        'help_edit': ''
    }
}


class StructureProcessor(base.Processor):

    name = 'structure'
    RESULT_TYPES = RESULTS

    def __init__(self, fail_fast=False, transform=False, report_limit=1000,
                 row_limit=30000, ignore_empty_rows=False,
                 ignore_duplicate_rows=False, ignore_defective_rows=False,
                 ignore_empty_columns=False, ignore_duplicate_columns=False,
                 ignore_headerless_columns=False, empty_strings=None,
                 report_stream=None, report=None, result_level='error',
                 **kwargs):

        # TODO: `self.seen` should be maintained in a file or something
        # TODO: Check for empty columns

        super(StructureProcessor, self).__init__(
            fail_fast=fail_fast, transform=transform, report_limit=report_limit,
            row_limit=row_limit, report_stream=report_stream, report=report,
            result_level=result_level)

        self.ignore_empty_rows = ignore_empty_rows
        self.ignore_duplicate_rows = ignore_duplicate_rows
        self.ignore_defective_rows = ignore_defective_rows
        self.ignore_empty_columns = ignore_empty_columns
        self.ignore_duplicate_columns = ignore_duplicate_columns
        self.ignore_headerless_columns = ignore_headerless_columns
        self.empty_strings = empty_strings or ('',)
        self.seen = {}

    def run_header(self, headers, header_index=0):

        valid = True

        # check for headerless columns
        if not self.ignore_headerless_columns:
            for index, header in enumerate(headers):
                if header in self.empty_strings:

                    valid = False
                    _type = RESULTS['empty_header']
                    entry = self.make_entry(
                        self.name,
                        self.RESULT_CATEGORY_HEADER,
                        self.RESULT_LEVEL_ERROR,
                        _type['msg'].format(index),
                        _type['id'],
                        _type['name'],
                        headers,
                        header_index,
                        self.RESULT_HEADER_ROW_NAME,
                        index,
                        header
                    )

                    self.report.write(entry)
                    if self.fail_fast:
                        return valid, headers

        # check for duplicate columns
        if not self.ignore_duplicate_columns:
            if len(set(headers)) != len(headers):

                valid = False
                dupes = [(index, header) for index, header in
                         enumerate(headers) if
                         header.count(header) > 1]
                _type = RESULTS['duplicate_header']

                for dupe in dupes:
                    entry = self.make_entry(
                        self.name,
                        self.RESULT_CATEGORY_HEADER,
                        self.RESULT_LEVEL_ERROR,
                        _type['msg'].format(dupe[0]),
                        _type['id'],
                        _type['name'],
                        headers,
                        header_index,
                        self.RESULT_HEADER_ROW_NAME,
                        dupe[0],
                        dupe[1]
                    )

                    self.report.write(entry)
                    if self.fail_fast:
                        return valid, headers

        return valid, headers

    def run_row(self, headers, index, row):

        valid, is_dupe, is_empty, is_defective = True, False, False, False
        row_name = self.get_row_id(row, headers)

        # check if row is duplicate
        if not self.ignore_duplicate_rows:

            _rep = hash(frozenset(row))

            if _rep in self.seen:

                # don't keep writing results for totally empty rows
                if set(row) == set(['']):
                    pass
                else:
                    previous_instances = copy.deepcopy(self.seen[_rep])
                    self.seen[_rep].append(index)
                    valid = False
                    is_dupe = True
                    _type = RESULTS['duplicate_row']
                    entry = self.make_entry(
                        self.name,
                        self.RESULT_CATEGORY_ROW,
                        self.RESULT_LEVEL_ERROR,
                        _type['msg'].format(index, previous_instances),
                        _type['id'],
                        _type['name'],
                        row,
                        index,
                        row_name
                    )

                    self.report.write(entry)
                    if self.fail_fast:
                        return valid, headers, index, row

            else:
                self.seen[_rep] = [index]

        # check if row is empty
        if not self.ignore_empty_rows:
            as_set = set(row)
            if len(as_set) == 1 and \
                    set(self.empty_strings).intersection(as_set):

                valid = False
                is_empty = True
                _type = RESULTS['empty_row']
                entry = self.make_entry(
                    self.name,
                    self.RESULT_CATEGORY_ROW,
                    self.RESULT_LEVEL_ERROR,
                    _type['msg'].format(index),
                    _type['id'],
                    _type['name'],
                    row,
                    index,
                    row_name
                )

                self.report.write(entry)
                if self.fail_fast:
                    return valid, headers, index, row

        # check if row is defective
        if not self.ignore_defective_rows:
            if len(headers) < len(row):

                valid = False
                is_defective = True
                _type = RESULTS['defective_row']
                entry = self.make_entry(
                    self.name,
                    self.RESULT_CATEGORY_ROW,
                    self.RESULT_LEVEL_ERROR,
                    _type['msg'].format(index),
                    _type['id'],
                    _type['name'],
                    row,
                    index,
                    row_name
                )

                self.report.write(entry)
                if self.fail_fast:
                    return valid, headers, index, row

            elif len(headers) < len(row):

                valid = False
                is_defective = True
                _type = RESULTS['defective_row']
                entry = self.make_entry(
                    self.name,
                    self.RESULT_CATEGORY_ROW,
                    self.RESULT_LEVEL_ERROR,
                    _type['msg'].format(index),
                    _type['id'],
                    _type['name'],
                    row,
                    index,
                    row_name
                )

                self.report.write(entry)
                if self.fail_fast:
                    return valid, headers, index, row

        if self.transform and any([is_dupe, is_empty, is_defective]):
            row = None

        return valid, headers, index, row



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys
import unittest
import click
import coverage


#CLI_DIR = os.path.abspath(os.path.dirname(__file__))
#REPO_DIR = os.path.abspath(os.path.dirname(os.path.dirname(CLI_DIR)))
#sys.path.insert(1, REPO_DIR)
from goodtables.pipeline import Pipeline


@click.group()
def cli():
    """The entry point into the CLI."""


@cli.command()
@click.argument('data')
@click.option('--schema')
@click.option('--format', default='csv', type=click.Choice(['csv', 'excel', 'json']))
@click.option('--dry_run', is_flag=True)
@click.option('--fail_fast', is_flag=True)
@click.option('--row_limit', default=20000, type=int)
@click.option('--report_limit', default=1000, type=int)
def validate(data, schema, format, dry_run, fail_fast, row_limit, report_limit):
    """Run a pipeline."""

    options = {}
    if schema:
        options['schema'] = {'schema': schema}

    pipeline = Pipeline(data, format=format, fail_fast=fail_fast,
                        dry_run=dry_run, row_limit=row_limit,
                        report_limit=report_limit)
    valid, report = pipeline.run()

    if valid:
        click.echo('The data source is valid.\n')
    else:
        click.echo('The data source is not valid.\n')
    click.echo('Following is a report::\n')
    click.echo(report)


if __name__ == '__main__':
    cli()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import io
import codecs
import datetime
import chardet
import xlrd
from .. import compat


class DataTable(object):

    """Convert a data source into a formatted (csv, json) utf-8 text stream."""

    REMOTE_SCHEMES = ('http', 'https', 'ftp', 'ftps')
    DEFAULT_ENCODING = 'utf-8'
    FORMATS = ('csv', 'excel', 'json')

    def __init__(self, data_source, headers=None, format='csv',
                 encoding=None, header_index=0, excel_sheet_index=0):

        self.openfiles = []
        self.data_source = data_source
        self.passed_headers = headers
        self.format = format
        self.encoding = encoding
        self.header_index = header_index
        self.excel_sheet_index = excel_sheet_index
        self.stream = self.to_textstream(self.data_source)
        self.headers, self.values = self.extract(self.passed_headers)

    def replay(self):
        """Replay the stream."""

        if self.stream.seekable():
            self.stream.seek(0)
        else:
            self.stream = self.to_textstream(self.data_source)

        self.headers, self.values = self.extract(self.passed_headers)
        return self.headers, self.values

    def extract(self, headers=None):
        """Extract headers and values from the data stream."""
        headers = headers or self.get_headers(self.stream)
        values = compat.csv_reader(self.stream)
        return headers, values

    def to_dict(self):
        raise NotImplementedError

    def to_csv(self, filepath=None):
        raise NotImplementedError

    def to_json(self, filepath=None):
        raise NotImplementedError

    def to_textstream(self, data_source):

        """Return a text stream from data_source.

        Args:
            data_source: An interface to data.
                Can be a stream, a string, a file path, or a URL path.

        Returns:
            A `self.DEFAULT_ENCODING` encoded text stream (utf-8).

        """

        # textstream = open('tmp.txt', mode='w+t', encoding=self.DEFAULT_ENCODING)
        textstream = io.TextIOWrapper(io.BufferedRandom(io.BytesIO()), encoding=self.DEFAULT_ENCODING)
        self.openfiles.append(textstream)

        if self.format in ('excel', 'json'):
            format_handler = getattr(self, '{0}_data_source'.format(self.format))
            data_source = format_handler(data_source)

        if isinstance(data_source, io.IOBase):

            if isinstance(data_source, io.TextIOBase):

                # if not data_source.encoding == self.DEFAULT_ENCODING:
                #    return
                return data_source

            else:

                stream_encoding = self._detect_stream_encoding(data_source)
                textstream = self._decode_to_textstream(data_source, stream_encoding, textstream)

                return textstream

        elif compat.parse.urlparse(data_source).scheme in self.REMOTE_SCHEMES:

            stream = self._stream_from_url(data_source)
            stream_encoding = self._detect_stream_encoding(stream)
            textstream = self._decode_to_textstream(stream, stream_encoding, textstream)

            return textstream

        elif isinstance(data_source, compat.str) and not \
                os.path.exists(data_source):

            stream_encoding = self._detect_stream_encoding(data_source)
            textstream = self._decode_to_textstream(data_source, stream_encoding, textstream)

            return textstream

        else:

            stream = io.open(data_source, mode='r+b')
            self.openfiles.append(stream)
            stream_encoding = self._detect_stream_encoding(stream)
            textstream = self._decode_to_textstream(stream, stream_encoding, textstream)

            return textstream

    def excel_data_source(self, data_source):
        """Get a data_source out of an Excel file."""

        # TODO: Need to flesh out this implementation quite a bit. See messytables

        if compat.parse.urlparse(data_source).scheme in self.REMOTE_SCHEMES:
            instream = compat.urlopen(data_source).read()
            workbook = xlrd.open_workbook(file_contents=instream)
        else:
            workbook = xlrd.open_workbook(data_source)

        out = io.TextIOWrapper(io.BufferedRandom(io.BytesIO()), encoding=self.DEFAULT_ENCODING)

        sheet = workbook.sheet_by_index(self.excel_sheet_index)
        row_count = sheet.nrows

        for row_index in range(row_count):

            values = []

            for cell in sheet.row(row_index):

                # TODO: this is very naive force to string
                if cell.ctype == 3:
                    value = datetime.datetime(
                        *xlrd.xldate_as_tuple(cell.value, sheet.book.datemode)
                    ).isoformat()
                else:
                    value = cell.value

                values.append(compat.str(cell.value))
                _data = ','.join(values)

            out.write('{0}\n'.format(_data))

        out.seek(0)
        return out

    def get_headers(self, stream):
        """Get headers from stream."""

        for index, line in enumerate(stream):
            if index == self.header_index:
                headers = line.rstrip('\n').split(',')
                break

        headers = [h.strip() for h in headers]
        return headers

    def _stream_from_url(self, url):
        """Return a seekable and readable stream from a URL."""

        stream = io.BufferedRandom(io.BytesIO())
        stream.write(compat.urlopen(url).read())
        stream.seek(0)

        return stream

    def _detect_stream_encoding(self, stream):
        """Return best guess at encoding of stream."""

        sample_length = 10000

        if self.encoding:
            return self.encoding

        if isinstance(stream, compat.str):
            if isinstance(stream, compat.bytes):
                sample = stream[:sample_length]
            else:
                sample = compat.to_bytes(stream)[:sample_length]
        else:
            sample = stream.read(sample_length)
            stream.seek(0)

        encoding = chardet.detect(sample)['encoding']

        return encoding

    def _decode_to_textstream(self, stream, encoding, textstream):
        """Return a textstream in `self.DEFAULT_ENCODING`"""

        if isinstance(stream, compat.str):
            _stream = io.StringIO()
            _stream.write(stream)
            stream = _stream
            stream.seek(0)
        else:
            stream = codecs.iterdecode(stream, encoding)

        for line in stream:
            recoded = line.encode(self.DEFAULT_ENCODING).decode(self.DEFAULT_ENCODING)
            textstream.write(recoded)

        textstream.seek(0)

        return textstream



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys
import tempfile
import io
import csv


_ver = sys.version_info
is_py2 = (_ver[0] == 2)
is_py3 = (_ver[0] == 3)
is_py33 = (is_py3 and _ver[1] == 3)
is_py34 = (is_py3 and _ver[1] == 4)
is_py27 = (is_py2 and _ver[1] == 7)


if is_py2:
    import urlparse as parse
    from urllib2 import urlopen
    builtin_str = str
    bytes = str
    str = unicode
    basestring = basestring
    numeric_types = (int, long, float)

    def csv_reader(data, dialect=csv.excel, **kwargs):
        """Read text stream (unicode on Py2.7) as CSV."""

        def iterenc_utf8(data):
            for line in data:
                yield line.encode('utf-8')

        reader = csv.reader(iterenc_utf8(data), dialect=dialect, **kwargs)
        for row in reader:
            yield [str(cell, 'utf-8') for cell in row]


elif is_py3:
    from urllib import parse
    from urllib.request import urlopen
    csv_reader = csv.reader
    builtin_str = str
    str = str
    bytes = bytes
    basestring = (str, bytes)
    numeric_types = (int, float)


def to_bytes(str):
    """Convert a text string to a byte string"""
    return str.encode('utf-8')


def to_builtin_str(str):
    """Convert a text string to the built-in `str` on the runtime."""
    if is_py2:
        return str.encode('utf-8')
    else:
        return str


def NamedTemporaryFile(mode='w+t', encoding='utf-8', **kwargs):
    """Return a NamedTemporaryFile for the appropriate runtime."""

    if is_py2:
        return tempfile.NamedTemporaryFile(mode=mode, **kwargs)

    elif is_py3:
        return tempfile.NamedTemporaryFile(mode=mode, encoding=encoding,
                                           **kwargs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import io
import tellme
from ..utilities import data_table, helpers
from .. import exceptions


class Processor(object):

    """Base Processor class. Processor implementations should inherit."""

    name = None
    ROW_LIMIT_MAX = 30000
    REPORT_LIMIT_MAX = 1000
    RESULT_CATEGORY_HEADER = 'header'
    RESULT_CATEGORY_ROW = 'row'
    RESULT_LEVEL_ERROR = 'error'
    RESULT_LEVEL_WARNING = 'warning'
    RESULT_LEVEL_INFO = 'info'
    RESULT_HEADER_ROW_NAME = 'headers'

    def __init__(self, fail_fast=False, transform=False, row_limit=30000,
                 header_index=0, report=None, report_limit=1000,
                 report_stream=None, result_level='error'):

        self.name = self.name or self.__class__.__name__.lower()
        self.fail_fast = fail_fast
        self.transform = transform
        self.row_limit = self.get_row_limit(row_limit)
        self.report_limit = self.get_report_limit(report_limit)
        self.header_index = header_index
        self.row_count = None
        self.result_level = result_level

        if report is None:
            if report_stream:
                report_stream_tests = [isinstance(report_stream, io.TextIOBase),
                                       report_stream.writable(),
                                       report_stream.seekable()]

                if not all(report_stream_tests):
                    _msg = '`report_stream` must be a seekable and writable text stream.'
                    raise exceptions.ProcessorBuildError(_msg)

                report_backend = 'client'
            else:
                report_backend = 'yaml'

            report_options = {
                'schema': helpers.report_schema,
                'backend': report_backend,
                'client_stream': report_stream,
                'limit': report_limit
            }

            self.report = tellme.Report(self.name, **report_options)
        else:
            self.report = report

    def get_row_limit(self, passed_limit):
        """Return the row limit, locked to an upper limit."""

        if passed_limit > self.ROW_LIMIT_MAX:
            return self.ROW_LIMIT_MAX
        else:
            return passed_limit

    def get_report_limit(self, passed_limit):
        """Return the report_limit, locked to an upper limit."""

        if passed_limit > self.REPORT_LIMIT_MAX:
            return self.REPORT_LIMIT_MAX
        else:
            return passed_limit

    def get_row_id(self, headers, row):
        """Return an identifying for a row, or None if not able to detect."""
        ids = ('id', '_id')
        for k, v in zip(headers, row):
            if k in ids:
                return v
        return ''

    def make_entry(self, processor, result_category, result_level, result_message,
                   result_id, result_name, result_context, row_index=None, row_name='',
                   column_index=None, column_name=''):
        """Return a report entry."""

        return {
            'processor': processor,
            'result_category': result_category,
            'result_level': result_level,
            'result_message': result_message,
            'result_id': result_id,
            'result_name': result_name,
            'result_context': result_context,
            'row_index': row_index,
            'row_name': row_name,
            'column_index': column_index,
            'column_name': column_name
        }

    def run(self, data_source, headers=None, encoding=None, is_table=False):

        """Run this processor on data_source.

        Args:
            data_source: the data source as string, URL, filepath or stream
            headers: pass headers explicitly to the DataTable constructor
            is_table: if True, data_source is a DataTable instance

        Returns:
            a tuple of `valid, report, data_table`
            valid: boolean indicating if the data is valid
            report: instance of reporter.Report
            data: a data_table.DataTable object containing the output data

        """

        def _run_valid(process_valid, run_valid):
            """Set/maintain the valid state of the run."""
            if not process_valid and run_valid:
                return False
            return run_valid

        valid = True
        openfiles = []

        if is_table:
            data = data_source
        else:
            data = data_table.DataTable(data_source, headers=headers,
                                        encoding=encoding,
                                        header_index=self.header_index)
            openfiles.extend(data.openfiles)

        # pre_run
        if hasattr(self, 'pre_run'):
            _valid, data = self.pre_run(data)
            valid = _run_valid(_valid, valid)
            if not _valid and self.fail_fast:
                return valid, self.report, data

        # run_header
        if hasattr(self, 'run_header'):
            _valid, data.headers = self.run_header(data.headers,
                                                   header_index=self.header_index)
            valid = _run_valid(_valid, valid)
            if not _valid and self.fail_fast:
                return valid, self.report, data

        # run_row
        if hasattr(self, 'run_row'):
            # TODO: on transform, create a new stream out of returned rows
            for index, row in enumerate(data.values):
                if index < self.row_limit:
                    # keep a count of rows processed
                    self.row_count = (index + 1)
                    _valid, data.headers, index, row = self.run_row(data.headers, index, row)
                    valid = _run_valid(_valid, valid)
                    if not _valid and self.fail_fast:
                        return valid, self.report, data

        # post_run
        if hasattr(self, 'post_run'):
            _valid, data = self.post_run(data)
            valid = _run_valid(_valid, valid)
            if not _valid and self.fail_fast:
                return valid, self.report, data

        for f in openfiles:
            f.close()

        return valid, self.report, data



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


class InvalidSpec(Exception):
    pass


class PipelineBuildError(Exception):

    def __init__(self, msg=None):
        self.msg = msg or 'Could not build the pipeline.'


class ProcessorBuildError(Exception):

    def __init__(self, msg=None):
        self.msg = msg or 'Could not build the processor.'



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys
import io
from setuptools import setup, find_packages


_ver = sys.version_info
is_py2 = (_ver[0] == 2)
is_py3 = (_ver[0] == 3)

dependencies = [
    'Click>=3.3',
    'jsonschema>=2.4.0',
    'chardet>=2.3.0',
    'xlrd>=0.9.3',
    'tellme>=0.1.6',
    'jtskit>=0.2.1'
]

setup(
    name='goodtables',
    description='Validate and process tabular data.',
    long_description='Validate and process tabular data.',
    version='0.3.5',
    author='Open Knowledge Foundation',
    author_email='info@okfn.org',
    url='http://okfn.org',
    license='MIT',
    packages=find_packages(exclude=['examples', 'tests']),
    include_package_data=True,
    zip_safe=False,
    install_requires=dependencies,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    entry_points={
        'console_scripts': [
            'goodtables = goodtables.cli.main:cli',
        ]
    },
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import io
import json
import requests
from .. import compat


def builtin_processors():
    """Return dict of public builtin processors. Avoids circular import."""
    from .. import processors
    return {
        processors.StructureProcessor.name: processors.StructureProcessor,
        processors.SchemaProcessor.name: processors.SchemaProcessor
    }


DEFAULT_PIPELINE = ('structure',)
REMOTE_SCHEMES = ('http', 'https', 'ftp', 'ftps')


# a schema for the reporter.Report() instances used by validators
report_schema = {
    'processor': {'type': compat.str},
    'result_category': {'type': compat.str},
    'result_level': {'type': compat.str},
    'result_message': {'type': compat.str},
    'result_id': {'type': compat.str},
    'result_name': {'type': compat.str},
    'result_context': {'type': list},
    'row_index': {'type': (int, type(None))},
    'row_name': {'type': compat.str},
    'column_index': {'type': (int, type(None))},
    'column_name': {'type': compat.str}
}


def get_report_result_types():
    """Return a list of all builtin result types."""

    result_types = []

    for processor in builtin_processors().values():
        result_types.extend([r[1] for r in
                             tuple(processor.RESULT_TYPES.items())])

    return result_types

def load_json_source(source):

    """Load a JSON source, from string, URL or buffer,  into a Python type."""

    if source is None:
        return None

    elif isinstance(source, (dict, list)):
        # the source has already been loaded
        return source

    elif compat.parse.urlparse(source).scheme in REMOTE_SCHEMES:
        return json.loads(requests.get(source).text)

    elif isinstance(source, compat.str) and not os.path.exists(source):
        return json.loads(source)

    else:
        with io.open(source, encoding='utf-8') as stream:
            source = json.load(io.open(source, encoding='utf-8'))
        return source
